import json
import re
import time

from six.moves import queue
import handler
import osvcd_shared as shared
import rcExceptions as ex
from rcUtilities import bdecode, normalize_jsonpath, is_string
from rcGlobalEnv import rcEnv
from jsonpath_ng import jsonpath
from jsonpath_ng.ext import parse
from converters import convert_boolean


OPERATORS = (">=", "<=", "=", ">", "<", "~", " in ")

class Handler(handler.Handler):
    """
    Wait <duration> for <condition> to become true.
    The <duration> is capped to 30 seconds.
    Upon timeout, it is up to the caller to re-submit the request until the condition becomes true.
    """
    routes = (
        ("GET", "wait"),
    )
    prototype = [
        {
            "name": "condition",
            "required": True,
            "format": "string",
            "desc": "An condition expressed as <jsonpath><operator><value>. The jsonpath is looked up in the daemon status dataset. Supported operators are %s." % " ".join(OPERATORS),
        },
        {
            "name": "duration",
            "required": True,
            "format": "duration",
            "desc": "An condition expressed as <jsonpath><operator><value>. The jsonpath is looked up in the daemon status dataset. Supported operators are = > < >= <= ~.",
        },
    ]
    access = {
        "roles": ["guest"],
        "namespaces": "ANY",
    }

    def action(self, nodename, thr=None, stream_id=None, **kwargs):
        thr.selector = "**"
        options = self.parse_options(kwargs)
        timeout = time.time() + options.duration
        if not thr.event_queue:
            thr.event_queue = queue.Queue()
        if not thr in thr.parent.events_clients:
            thr.parent.events_clients.append(thr)
        neg, jsonpath_expr, oper, val = self.parse_condition(options.condition)
        if neg ^ self.match(jsonpath_expr, oper, val, {"kind": "patch"}):
            return {"status": 0, "data": {"satisfied": True, "duration": options.duration, "elapsed": 0}}
        while True:
            left = timeout - time.time()
            try:
                msg = thr.event_queue.get(True, left)
            except queue.Empty:
                return {"status": 1, "data": {"satisfied": False, "duration": options.duration, "elapsed": options.duration-left}}
            if neg ^ self.match(jsonpath_expr, oper, val, msg):
                return {"status": 0, "data": {"satisfied": True, "duration": options.duration, "elapsed": options.duration-left}}

    def parse_condition(self, condition):
        oper = None
        val = None

        if condition[0] == "!":
            path = condition[1:]
            neg = True
        else:
            path = condition
            neg = False

        for op in OPERATORS:
            idx = path.rfind(op)
            if idx < 0:
                continue
            val = path[idx+len(op):].strip()
            path = path[:idx].strip()
            oper = op
            if op == "~":
                if not val.startswith(".*") and not val.startswith("^"):
                    val = ".*" + val
                if not val.endswith(".*") and not val.endswith("$"):
                    val = val + ".*"
            break

        path = normalize_jsonpath(path)
        try:
            jsonpath_expr = parse(path)
        except Exception as exc:
            raise ex.excError(exc)

        return neg, jsonpath_expr, oper, val

    def eval_condition(self, jsonpath_expr, oper, val, data):
        for match in jsonpath_expr.find(data):
            if oper is None:
                if match.value:
                    return True
                else:
                    continue
            obj_class = type(match.value)
            try:
                if obj_class == bool:
                    val = convert_boolean(val)
                else:
                    val = obj_class(val)
            except Exception as exc:
                raise ex.excError("can not convert to a common type")
            if oper is None:
                if match.value:
                    return True
            if oper == "=":
                if match.value == val:
                    return True
            elif oper == ">":
                if match.value > val:
                    return True
            elif oper == "<":
                if match.value < val:
                    return True
            elif oper == ">=":
                if match.value >= val:
                    return True
            elif oper == "<=":
                if match.value <= val:
                    return True
            elif is_string(match.value) and oper == "~":
                if re.match(val, match.value):
                    return True
            elif oper == " in ":
                try:
                    l = json.loads(val)
                except:
                    l = val.split(",")
                if match.value in l:
                    return True
        return False

    def match(self, jsonpath_expr, oper, val, msg):
        kind = msg.get("kind")
        if kind == "patch":
            if self.eval_condition(jsonpath_expr, oper, val, shared.LAST_DAEMON_STATUS):
                return True
        elif kind == "event":
            if self.eval_condition(jsonpath_expr, oper, val, msg):
                return True
        return False

